// cConio.cpp
//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).

#import "cConio3d.hpp"

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <string.h>


void printChar3d(char c, SCALAR size, cVector p)
{
  INTEG n = c;
  SCALAR s = 1.0 / 32.0;
  SCALAR t = -1.0 / 8.0;
  SCALAR e = (n % 32) * s;
  SCALAR f = (n / 32) * t;

  glBegin(GL_POLYGON);

    glTexCoord2f(e, f);
    glVertex3f(p.mV[X],      p.mV[Y],      p.mV[Z]);

    glTexCoord2f(e+s, f);
    glVertex3f(p.mV[X]+size, p.mV[Y],      p.mV[Z]);

    glTexCoord2f(e+s, f+t);
    glVertex3f(p.mV[X]+size, p.mV[Y]-size, p.mV[Z]);

    glTexCoord2f(e, f+t);
    glVertex3f(p.mV[X],      p.mV[Y]-size, p.mV[Z]);

  glEnd();
}


void print3d(char *str, SCALAR size, cVector p, DWORD tex, cMatrix camera)
{
  cMatrix cm = camera;
  glLoadMatrixf(cm.mM);
  glBindTexture(GL_TEXTURE_2D, tex);
  INTEG i, n = strlen(str);
  cVector v = p;
  for (i=0; i<n; i++)
  {
    printChar3d(str[i], size, v);
    v.mV[X] += size;
  };
}







cConio3d::cConio3d()
{
  mDisplay = NULL;
  Resize(30, 25, 0.02);
  mBackColor[0].Set(0.0, 0.30, 1.0, 0.85);
  mBackColor[1].Set(0.0, 0.15, 0.5, 0.85);
  mBackColor[2].Set(0.0, 0.0, 0.0, 0.85);
  mBackColor[3].Set(0.0, 0.15, 0.5, 0.85);
  mFrontColor.Set(0.9, 0.9, 0.9, 1.0);
  mCursorColor.Set(0.0, 0.6, 0.0, 0.6);
};


void cConio3d::Resize(INTEG w, INTEG h, SCALAR size)
{
  mW = w;
  mH = h;
  mSize = size;
  mPos = 0;
  delete [] mDisplay;
  mDisplay = new char[mW*mH];
  INTEG i;
  for (i=0; i<mW*mH; i++) mDisplay[i] = ' ';
}


void cConio3d::SetFont(DWORD font)
{
  mFont = font;
}


void cConio3d::Scroll()
{
  INTEG i;
  memcpy(&mDisplay[0], &mDisplay[mW], mW*mH-mW);
  for (i=0; i<mW; i++) mDisplay[mW*mH-mW+i] = ' ';
}


void cConio3d::InsertChar(char c)
{
  INTEG i;
  if (c == '\n') c = 13;

  if (mPos >= mW*mH ) return;
  if (c == 8) {   // backspace
    mPos--;
    if (mPos < 0)
      mPos = 0;
    else
      mDisplay[mPos] = ' ';
  } else if (c == 9) {
    Insert("     ");    
  } else if (c == 13) {  // return
    INTEG rest = mW - (mPos % mW);
    for (i=0; i<rest; i++) InsertChar(' ');
  } else {  // insert the character and move cursor
    mDisplay[mPos] = c;
    mPos++;
    if (mPos >= mW*mH) {
      Scroll();
      mPos = mW*mH-mW;
    };
  };
}


void cConio3d::Insert(char *s)
{
  INTEG i, n = strlen(s);
  for (i=0; i<n; i++) InsertChar(s[i]);
};


void cConio3d::Draw(cMatrix camera)
{
  cMatrix m;
  cMatrix n = camera;
  cMatrix cm = mPose.GetObjectMatrix();
  m.MovRel(mCenter.mV[X], mCenter.mV[Y], mCenter.mV[Z]);
  m.Mul(cm);
  m.Mul(n);
  glLoadMatrixf(m.mM);
  
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
    glColor4fv(mBackColor[0].mV);
    glVertex3f(0.0-mSize,  0.0+mSize, -0.01);
    glColor4fv(mBackColor[1].mV);
    glVertex3f(mSize*mW+mSize,  0.0+mSize, -0.01);
    glColor4fv(mBackColor[2].mV);
    glVertex3f(mSize*mW+mSize, -mSize*mH-mSize, -0.01);
    glColor4fv(mBackColor[3].mV);
    glVertex3f(0.0-mSize, -mSize*mH-mSize, -0.01);
  glEnd();
  glEnable(GL_TEXTURE_2D);
  
  glColor4fv(mFrontColor.mV);
  glBindTexture(GL_TEXTURE_2D, mFont);  
  cVector v;
  INTEG i;
  for (i=0; i<mW*mH; i++)
  {
    printChar3d(mDisplay[i], mSize, v);
    if ((i>0) && ((i+1)%(mW)==0)) {
      v.mV[X]  = 0;
      v.mV[Y] -= mSize;
    } else {
      v.mV[X] += mSize;
    };
  };

  SCALAR h = -(mPos / mW)*mSize;
  SCALAR w = (mPos % mW)*mSize;
  glDisable(GL_TEXTURE_2D);
  glColor4fv(mCursorColor.mV);
  glBegin(GL_QUADS);
    glVertex3f(w,        h,       +0.01);
    glVertex3f(w+mSize,  h,       +0.01);
    glVertex3f(w+mSize,  h-mSize, +0.01);
    glVertex3f(w,        h-mSize, +0.01);
  glEnd();
  glEnable(GL_TEXTURE_2D);
}

