// cPlane.cpp

#import "cPlane.hpp"


void cPlane::Set(cVector v, SCALAR d)
{
  mV = v;
  mV.mV[D] = d;
}


void cPlane::Set(SCALAR a, SCALAR b, SCALAR c, SCALAR d)
{
  mV.Set(a, b, c);
  mV.mV[D] = d;
}


void cPlane::Set(cVector normal, cVector point)
{
  mV = normal;
  mV.Normalise();
  mV.mV[D] = calcD(mV.mV, point.mV);
}


SCALAR cPlane::Distance(cVector point)
{
  return ( calcP(mV.mV, point.mV, mV.mV[D]) );
}


BOOLEAN cPlane::Intersect(cVector pointA, cVector pointB, cVector &pointS)
{
  SCALAR d0 = calcP(mV.mV, pointA.mV, mV.mV[D]);
  SCALAR d1 = calcP(mV.mV, pointB.mV, mV.mV[D]);
  if ( xsgn(d0) == xsgn(d1) ) return 0; // sorry, no intersection!
  SCALAR dd = d0 / (d0 - d1);
  pointS.mV[X] = pointA.mV[X] - ((pointA.mV[X] - pointB.mV[X]) * dd);
  pointS.mV[Y] = pointA.mV[Y] - ((pointA.mV[Y] - pointB.mV[Y]) * dd);
  pointS.mV[Z] = pointA.mV[Z] - ((pointA.mV[Z] - pointB.mV[Z]) * dd);
  return 1;
}


cVector cPlane::Intersect(cRay ray)
{
  cVector v = ray.mO, r = ray.mD;
  SCALAR d = calcT(mV.mV, v.mV, r.mV, mV.mV[D]);
  r.Mul(d);
  v.Add(r);
  return v;
}


