// cSphere.cpp

#import "cSphere.hpp"

#import "cRay.hpp"


cSphere::cSphere()
{
  mR = 0;
  mR2 = 0;
}


void cSphere::Set(cVector *vList, INTEG n)
{
  mV.CenterOf(vList, n);
  mR = lineDistance(mV.mV, vList[mV.Furthest(vList, n)].mV);
  mR2 = xsqr(mR);
}


BOOLEAN cSphere::Hit(cSphere other)
{
  SCALAR d = lineDistance2(mV.mV, other.mV.mV);
  if (d <= (mR2 + other.mR2)) return 1;
  return 0;
}


BOOLEAN cSphere::Inside(cVector point)
{
  if (lineDistance(mV.mV, point.mV) <= mR) return 1;
  return 0;
}


BOOLEAN cSphere::Hit(cRay ray, cVector &point)
{
  cVector H, N;
  SCALAR  a, b, c, d, s1, s2, s;

  H.Set(ray.mO, mV);  // origin to center

  a = xsqr(ray.mD.mV[X]) + xsqr(ray.mD.mV[Y]) + xsqr(ray.mD.mV[Z]);
  b = 2.0*(H.mV[X]*ray.mD.mV[X] + H.mV[Y]*ray.mD.mV[Y] + H.mV[Z]*ray.mD.mV[Z]);
  c = xsqr(H.mV[X]) + xsqr(H.mV[Y]) + xsqr(H.mV[Z]) - xsqr(mR);
  d = xsqr(b) - 4.0*a*c; 
 
  if(d >= 0.0) {
    s1 = -b + sqrt(d) / (2*a);
    s2 = -b - sqrt(d) / (2*a);
    if((s1 > 0) || (s2 > 0)) {   // One of them is positive.
      if (s1 <= 0) s1 = s2;      // If s1 is negative then kick it.
      if (s2 <= 0) s2 = s1;      // Same here but the other way.
      s = xmin(s1, s2);          // The nearest (positive).
      point = ray.mD;
      point.Mul(s);              // Take s steps into given direction,
      point.Add(ray.mO);         // from ray origin.
		  N.Set(point, mV);          // Center to intersection (normal of sphere) - still unused.
      N.Normalise();
      return 1; 
    };
  };
  return 0; 
}


