// cSprite.cpp
//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//

#import "cSprite.hpp"
#import "cBody.hpp"

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL\gl.h>
#include <GL\glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include <stdlib.h>


#define distance2d(A, B)       (sqrt( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y]) ))


cSprite::cSprite()
{
  mNPnt = 0;
  mNPlg = 0;
  mLmb = NULL;
  mTex = NULL;
  mPnt = NULL;
  mPlg = NULL;
  mSel.mPnt[0] = 0;
  mSel.mPnt[1] = 0;
  mSel.mPnt[2] = 0;
  mSel.mSide = 0;
  mBody = NULL;
}


void cSprite::GetMem(INTEG pnt, INTEG plg)
{
  mNPnt = 0;
  mNPlg = 0;
  delete [] mLmb;
  delete [] mTex;
  delete [] mPnt;
  delete [] mPlg;
  mLmb = new INTEG[pnt];
  mTex = new cVector[pnt];
  mPnt = new cVector[pnt];
  mPlg = new cRawPolygon[plg];
}


void cSprite::Save(char *fname)
{
  FILE *f;
  INTEG i, a, b, c, d;
  f = fopen(fname, "w");
  if (f == NULL) return;
  fprintf(f, "HM_OPT\n");
  fprintf(f, "1.000000\n");
  fprintf(f, "%i\n", mNPnt);
  for (i=0; i<mNPnt; i++)
  {
    a = (INTEG) (mPnt[i].mV[X]*1000.0);
    b = (INTEG) (mPnt[i].mV[Y]*1000.0);
    c = (INTEG) (mPnt[i].mV[Z]*1000.0);
    fprintf(f, "%i %i %i\n", a, b, c);
  };
  fprintf(f, "%i\n", mNPlg);
  for (i=0; i<mNPlg; i++)
  {
    a = mPlg[i].mPnt[0];
    b = mPlg[i].mPnt[1];
    c = mPlg[i].mPnt[2];
    d = mPlg[i].mSide + 1;
    fprintf(f, "%i %i %i %i\n", a, b, c, d);
  };
  fprintf(f, "%i\n", mNPnt);
  for (i=0; i<mNPnt; i++)
  {
    a = (INTEG) (mPnt[i].mV[X]*1000.0);
    b = (INTEG) (mPnt[i].mV[Y]*1000.0);
    c = (INTEG) (mPnt[i].mV[Z]*1000.0);
    fprintf(f, "%i %i %i\n", a, b, c);
  };
  fprintf(f, "%i\n", mNPnt);
  for (i=0; i<mNPnt; i++)
  {
    a = mLmb[i];
    fprintf(f, "%i\n", a);
  };
  fclose(f);
}


void cSprite::Load(char *fname)
{
  FILE *f;
  INTEG i, a, b, c, d, dummy;
  SCALAR scale;
  char bufs[80];
  f = fopen(fname, "r");
  if (f == NULL) return;
  fscanf(f, "%s\n", bufs);
  fscanf(f, "%f\n", &scale);
  fscanf(f, "%i\n", &mNPnt);
  for (i=0; i<mNPnt; i++)
  {
    fscanf(f, "%i %i %i\n", &a, &b, &c);
    mPnt[i].mV[X] = a / 1000.0;
    mPnt[i].mV[Y] = b / 1000.0;
    mPnt[i].mV[Z] = c / 1000.0;
  };
  fscanf(f, "%i\n", &mNPlg);
  for (i=0; i<mNPlg; i++)
  {
    fscanf(f, "%i %i %i %i\n", &a, &b, &c, &d);
    mPlg[i].mPnt[0] = a;
    mPlg[i].mPnt[1] = b;
    mPlg[i].mPnt[2] = c;
    mPlg[i].mSide   = d - 1;
  };
  fscanf(f, "%i\n", &dummy);
  for (i=0; i<dummy; i++)
  {
    fscanf(f, "%i %i %i\n", &a, &b, &c);
    mTex[i].mV[X] = a / 1000.0;
    mTex[i].mV[Y] = b / 1000.0;
    mTex[i].mV[Z] = c / 1000.0;
  };
  fscanf(f, "%i\n", &dummy);
  for (i=0; i<dummy; i++)
  {
    fscanf(f, "%i\n", &a);
    mLmb[i] = a;
  };
  fclose(f);
}


void cSprite::DragFront(cVector pos)
{
  if (mSel.mPnt[0]>=0)
  if (mPnt[mSel.mPnt[0]].mV[Z]>=0) {
    pos.mV[Z] = mPnt[mSel.mPnt[0]].mV[Z]; // don't change Z
    mPnt[mSel.mPnt[0]] = pos;
  };
}


void cSprite::DragBack(cVector pos)
{
  if (mSel.mPnt[0]>=0)
  if (mPnt[mSel.mPnt[0]].mV[Z]<=0) {
    pos.mV[Z] = mPnt[mSel.mPnt[0]].mV[Z]; // don't change Z
    mPnt[mSel.mPnt[0]] = pos;
  };
}


void cSprite::PushFront(cVector pos, SCALAR offset)
{
  INTEG i, a;
  double e, d;
  a = -1;
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] > 0)
  {
    e = distance2d(pos.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      a = i;
    };
  };
  if (a>=0) {
    mPnt[a].mV[Z] += offset;
    if (mPnt[a].mV[Z] < 0.01) mPnt[a].mV[Z] = 0.01;
  };
}


void cSprite::PushBack(cVector pos, SCALAR offset)
{
  INTEG i, a;
  double e, d;
  a = -1;
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] < 0)
  {
    e = distance2d(pos.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      a = i;
    };
  };
  if (a>=0) {
    mPnt[a].mV[Z] += offset;
    if (mPnt[a].mV[Z] > -0.01) mPnt[a].mV[Z] = -0.01;
  };
}


void cSprite::Draw(cMatrix camera)
{
  glLoadMatrixf(camera.mM);

  if (!mBody) return;

  INTEG i, m, n, o, u;
  cVector a, b, c;
  cVector *tpnt = new cVector[mNPnt];

  for (i=0; i<mNPnt; i++)
  {
    u = mLmb[i];
    tpnt[i] = mPnt[i];
    tpnt[i].Sub(mBody->mBones[u]->mUBase);
    mBody->mBones[u]->mLimb.Apply(tpnt[i]);
    tpnt[i].Add(mBody->mBones[u]->mUJoint);
  };

  glEnable(GL_TEXTURE_2D);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBindTexture(GL_TEXTURE_2D, mTexture[0]);
  for (i=0; i<mNPlg; i++) if (mPlg[i].mSide == +1)
  {
    m = mPlg[i].mPnt[0];
    n = mPlg[i].mPnt[1];
    o = mPlg[i].mPnt[2];
    a = tpnt[m];
    b = tpnt[n];
    c = tpnt[o];
    glBegin(GL_TRIANGLES);
      glTexCoord2f(0.5+mTex[m].mV[X], 0.5-mTex[m].mV[Y]);
      glVertex3fv(a.mV);
      glTexCoord2f(0.5+mTex[n].mV[X], 0.5-mTex[n].mV[Y]);
      glVertex3fv(b.mV);
      glTexCoord2f(0.5+mTex[o].mV[X], 0.5-mTex[o].mV[Y]);
      glVertex3fv(c.mV);
    glEnd();
  };

  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBindTexture(GL_TEXTURE_2D, mTexture[1]);
  for (i=0; i<mNPlg; i++) if (mPlg[i].mSide == -1)
  {
    m = mPlg[i].mPnt[0];
    n = mPlg[i].mPnt[1];
    o = mPlg[i].mPnt[2];
    a = tpnt[m];
    b = tpnt[n];
    c = tpnt[o];
    glBegin(GL_TRIANGLES);
      glTexCoord2f(0.5+mTex[m].mV[X], 0.5-mTex[m].mV[Y]);
      glVertex3fv(a.mV);
      glTexCoord2f(0.5+mTex[n].mV[X], 0.5-mTex[n].mV[Y]);
      glVertex3fv(b.mV);
      glTexCoord2f(0.5+mTex[o].mV[X], 0.5-mTex[o].mV[Y]);
      glVertex3fv(c.mV);
    glEnd();
  };
  delete [] tpnt;
}


void cSprite::Draw()
{
  INTEG i, m, n, o;
  cVector a, b, c;

  glDisable(GL_TEXTURE_2D);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  glColor4f(0.0, 1.0, 0.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] == 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glColor4f(1.0, 0.0, 0.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] > 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glColor4f(0.0, 0.0, 1.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] < 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glEnable(GL_TEXTURE_2D);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBindTexture(GL_TEXTURE_2D, mTexture[0]);
  for (i=0; i<mNPlg; i++) if (mPlg[i].mSide == +1)
  {
    m = mPlg[i].mPnt[0];
    n = mPlg[i].mPnt[1];
    o = mPlg[i].mPnt[2];
    a = mPnt[m];
    b = mPnt[n];
    c = mPnt[o];
    glBegin(GL_TRIANGLES);
      glTexCoord2f(0.5+mTex[m].mV[X], 0.5-mTex[m].mV[Y]);
      glVertex3fv(a.mV);
      glTexCoord2f(0.5+mTex[n].mV[X], 0.5-mTex[n].mV[Y]);
      glVertex3fv(b.mV);
      glTexCoord2f(0.5+mTex[o].mV[X], 0.5-mTex[o].mV[Y]);
      glVertex3fv(c.mV);
    glEnd();
  };

  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBindTexture(GL_TEXTURE_2D, mTexture[1]);
  for (i=0; i<mNPlg; i++) if (mPlg[i].mSide == -1)
  {
    m = mPlg[i].mPnt[0];
    n = mPlg[i].mPnt[1];
    o = mPlg[i].mPnt[2];
    a = mPnt[m];
    b = mPnt[n];
    c = mPnt[o];
    glBegin(GL_TRIANGLES);
      glTexCoord2f(0.5+mTex[m].mV[X], 0.5-mTex[m].mV[Y]);
      glVertex3fv(a.mV);
      glTexCoord2f(0.5+mTex[n].mV[X], 0.5-mTex[n].mV[Y]);
      glVertex3fv(b.mV);
      glTexCoord2f(0.5+mTex[o].mV[X], 0.5-mTex[o].mV[Y]);
      glVertex3fv(c.mV);
    glEnd();
  };

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


void cSprite::DrawFront()
{
  INTEG i, m, n, o;

  glDisable(GL_TEXTURE_2D);

  glColor4f(0.0, 1.0, 0.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] == 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glColor4f(1.0, 0.0, 0.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] > 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glColor4f(1.0, 0.2, 0.0, 1.0);
  for (i=0; i<mNPlg; i++) if (mPlg[i].mSide == +1)
  {
    m = mPlg[i].mPnt[0];
    n = mPlg[i].mPnt[1];
    o = mPlg[i].mPnt[2];
    glBegin(GL_TRIANGLES);
      glVertex3f(mPnt[m].mV[X], mPnt[m].mV[Y], mPnt[m].mV[Z]);
      glVertex3f(mPnt[n].mV[X], mPnt[n].mV[Y], mPnt[n].mV[Z]);
      glVertex3f(mPnt[o].mV[X], mPnt[o].mV[Y], mPnt[o].mV[Z]);
    glEnd();
  };

  glColor4f(1.0, 1.0, 1.0, 0.5);
  if (mSel.mSide == +1) {
    m = mSel.mPnt[0];
    n = mSel.mPnt[1];
    o = mSel.mPnt[2];
    glBegin(GL_TRIANGLES);
      glVertex3f(mPnt[m].mV[X], mPnt[m].mV[Y], mPnt[m].mV[Z]);
      glVertex3f(mPnt[n].mV[X], mPnt[n].mV[Y], mPnt[n].mV[Z]);
      glVertex3f(mPnt[o].mV[X], mPnt[o].mV[Y], mPnt[o].mV[Z]);
    glEnd();
  };
}


void cSprite::DrawBack()
{
  INTEG i, m, n, o;

  glDisable(GL_TEXTURE_2D);

  glColor4f(0.0, 1.0, 0.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] == 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glColor4f(0.0, 0.0, 1.0, 1.0);
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] < 0)
  {
    glBegin(GL_LINES);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y], mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X], mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };

  glColor4f(1.0, 0.2, 0.0, 1.0);
  for (i=0; i<mNPlg; i++) if (mPlg[i].mSide == -1)
  {
    m = mPlg[i].mPnt[0];
    n = mPlg[i].mPnt[1];
    o = mPlg[i].mPnt[2];
    glBegin(GL_TRIANGLES);
      glVertex3f(mPnt[m].mV[X], mPnt[m].mV[Y], mPnt[m].mV[Z]);
      glVertex3f(mPnt[n].mV[X], mPnt[n].mV[Y], mPnt[n].mV[Z]);
      glVertex3f(mPnt[o].mV[X], mPnt[o].mV[Y], mPnt[o].mV[Z]);
    glEnd();
  };

  glColor4f(1.0, 1.0, 1.0, 0.5);
  if (mSel.mSide == -1) {
    m = mSel.mPnt[0];
    n = mSel.mPnt[1];
    o = mSel.mPnt[2];
    glBegin(GL_TRIANGLES);
      glVertex3f(mPnt[m].mV[X], mPnt[m].mV[Y], mPnt[m].mV[Z]);
      glVertex3f(mPnt[n].mV[X], mPnt[n].mV[Y], mPnt[n].mV[Z]);
      glVertex3f(mPnt[o].mV[X], mPnt[o].mV[Y], mPnt[o].mV[Z]);
    glEnd();
  };
}


void cSprite::MarkFrontLimb(INTEG lmb)
{
  glDisable(GL_TEXTURE_2D);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glColor4f(0.0, 0.0, 1.0, 1.0);
  INTEG i;
  for (i=0; i<mNPnt; i++) if ((mPnt[i].mV[Z] >= 0) && (mLmb[i] == lmb))
  {
    glBegin(GL_QUADS);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


void cSprite::MarkBackLimb(INTEG lmb)
{
  glDisable(GL_TEXTURE_2D);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glColor4f(0.0, 0.0, 1.0, 1.0);
  INTEG i;
  for (i=0; i<mNPnt; i++) if ((mPnt[i].mV[Z] <= 0) && (mLmb[i] == lmb))
  {
    glBegin(GL_QUADS);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y]-0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]+0.01, mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
      glVertex3f(mPnt[i].mV[X]-0.01, mPnt[i].mV[Y]+0.01, mPnt[i].mV[Z]);
    glEnd();
  };
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


INTEG cSprite::SelectFront(cVector center)
{
//  if (mNPnt < 3) return;
  INTEG i, a, b, c;
  double e, d;
  a = b = c = -1;
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] >= 0)
  {
    e = distance2d(center.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      a = i;
    };
  };
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if (i != a) if (mPnt[i].mV[Z] >= 0)
  {
    e = distance2d(center.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      b = i;
    };
  };
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if ((i != a) && (i != b))  if (mPnt[i].mV[Z] >= 0)
  {
    e = distance2d(center.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      c = i;
    };
  };
  mSel.mPnt[0] = a;
  mSel.mPnt[1] = b;
  mSel.mPnt[2] = c;
  mSel.mSide = 0;
  if ((a>=0) && (b>=0) && (c>=0)) mSel.mSide = +1;
  return a;
}


INTEG cSprite::SelectBack(cVector center)
{
//  if (mNPnt < 3) return;
  INTEG i, a, b, c;
  double e, d;
  a = b = c = -1;
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if (mPnt[i].mV[Z] <= 0)
  {
    e = distance2d(center.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      a = i;
    };
  };
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if (i != a) if (mPnt[i].mV[Z] <= 0)
  {
    e = distance2d(center.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      b = i;
    };
  };
  d = 100000000.0;
  for (i=0; i<mNPnt; i++) if ((i != a) && (i != b))  if (mPnt[i].mV[Z] <= 0)
  {
    e = distance2d(center.mV, mPnt[i].mV);
    if (e<d) {
      d = e;
      c = i;
    };
  };
  mSel.mPnt[0] = a;
  mSel.mPnt[1] = b;
  mSel.mPnt[2] = c;
  mSel.mSide = 0;
  if ((a>=0) && (b>=0) && (c>=0)) mSel.mSide = -1;
  return a;
}


void cSprite::AddFrontPoly(cVector center)
{
  SelectFront(center);
  if (!mSel.mSide) return;
  mPlg[mNPlg] = mSel;
  mPlg[mNPlg].mSide = +1;
  mNPlg++;
}


void cSprite::AddBackPoly(cVector center)
{
  SelectBack(center);
  if (!mSel.mSide) return;
  mPlg[mNPlg] = mSel;
  mPlg[mNPlg].mSide = -1;
  mNPlg++;
}

