// cController.cpp
//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
//

#include "cController.h"
#include "cObject.h"

#include <cstdlib>
#include <cassert>

using namespace std;

#define debug !true

//--------------------------------


cController::cController()
{
  mDevice = NULL;

  clearState();
}


void cController::clearState()
{
  if (debug) printf("clearState()\n");
  //  overideState(WAIT, (char*) NULL);
}


void cController::overideState(int state, cObject *entity)
{
  if (debug) printf("overideState(%s)\n", getStateName(state));

  if (entity == NULL) return;
  if (entity->mFragged) return;

  state_struct* st;

  if (!mState.empty()) {
    st = mState.top();
    if (st->entity_name != NULL) delete st->entity_name;
    mState.pop();
    delete st;
  };

  st = new state_struct;
  st->state = state;
  st->entity = entity;
  if (entity->mName != NULL) st->entity_name = new string(*(entity->mName));
  mState.push(st);
}


void cController::overideState(int state, char* entityname)
{
  if (debug) printf("overideState(%s, %s)\n", getStateName(state), entityname);

  if (entityname == NULL) return;

  state_struct* st;

  if (!mState.empty())
  {
    st = mState.top();
    if (st->entity_name != NULL) delete st->entity_name;
    mState.pop();
    delete st;
  };

  st = new state_struct;
  st->state = state;
  st->entity = NULL;
  st->entity_name = new string(entityname);
  mState.push(st);
}


void cController::newState(int state, cObject *entity)
{
  if (debug) printf("newState(%s)\n", getStateName(state));

  if (entity == NULL) return;
  else if (entity->mFragged) return;

  state_struct* st;

  st = new state_struct;
  st->state = state;
  st->entity = entity;
  if (entity->mName != NULL) st->entity_name = new string(*(entity->mName));
  mState.push(st);
}


void cController::newState(int state, char* entityname)
{
  if (debug) printf("overideState(%s, %s)\n", getStateName(state), entityname);

  if (entityname == NULL) return;

  state_struct* st;

  st = new state_struct;
  st->state = state;
  st->entity = NULL;
  st->entity_name = new string(entityname);
  mState.push(st);
}


void cController::oldState()
{
  if (debug) printf("oldState()\n");

  // Is there any state to remove?
  if (mState.size() == 0) return;

  state_struct* st;

  // Remove state.
  st = mState.top();
  if (st->entity_name != NULL) delete st->entity_name;
  mState.pop();
  delete st;

  if (mState.size() > 0) {
    st = mState.top();
    st->entity = NULL; // Pointer might be invalid, search again.
  };
}


void cController::processSelf()
{
  
  if (mDevice == NULL) return;

  if (mState.empty()) {
    waitEnemy();
    return;
  };
  
  //if (mState.empty()) return;
  state_struct* st = mState.top();

  if (debug && mDevice->mName != NULL) printf("Processing: %s\n", mDevice->mName->c_str());
  if (debug && st->entity_name != NULL) printf("processSelf(#%i) = <%s>\n", mState.size(), st->entity_name->c_str());

  if ((st->entity == NULL) && st->entity_name != NULL)
  {
    // Search world for the entity name.
    cWorld* world = mDevice->mWorld;
    list<cObject*>* roles = world->filterByRole(NULL, (1UL<<cObject::MECH), false, -1);
    if (debug) world->printNames();
    if (debug) printf("|filter by role| = %i\n", roles->size());
    list<cObject*>* result = world->filterByName(mDevice, (char*) st->entity_name->c_str(), 1, roles);
    if (debug) printf("|filter by name| = %i\n", result->size());
    if (!result->empty()) {
        st->entity = (cObject*) result->front();
    } else {
        printf("no result of findName for <%s>\n", st->entity_name->c_str());
        if (debug) world->printNames(roles);
        delete result;
        delete roles;
        return;
    }
    delete result;
    delete roles;
  };

  // Set Target Destination Vector.
  if (st->entity != NULL) {
      mDevice->mTarget = st->entity;
  } else {
      mDevice->mTarget = NULL;
  }

  switch(st->state) {
    case WAIT: waitEnemy();      break;
    case ENGAGE: engageEnemy();    break;
    case TARGET: targetEnemy();    break;
    case ATTACK:  attackEnemy();    break;
    case FOLLOW: followLeader();   break;
    case FLEE: fleeEnemy();      break;
    case ROUTE: routeNav();       break;
    case PATROL: patrolArea();     break;
    default: report("Invalid State Request!\n");     break;
  };
}


void cController::waitEnemy()
{
  // Search world for any enemy in range.
  cObject* nearby = enemyNearby();
  if (nearby) {
      if (mDevice->mName != NULL && nearby->mName != NULL) {
          printf("> %s Targets %s\n", mDevice->mName->c_str(), nearby->mName->c_str());
      }
      newState(TARGET, nearby);
  }

  {
    do_idle();
  }
}


void cController::engageEnemy()
{
  if (inTargetRange() > 0.9)
  {
    if (mDevice->mName != NULL && mState.top()->entity->mName != NULL) {
        printf("> %s Targets %s\n", mDevice->mName->c_str(), mState.top()->entity->mName->c_str());
    }
    newState(TARGET, mState.top()->entity);
    return;
  };

  {
      do_moveTowards();
      do_aimAt();
  }
}


void cController::targetEnemy()
{
  if (inMeeleRange() > 0.9)
  {
    if (mDevice->mName != NULL && mState.top()->entity->mName != NULL) {
        printf("> %s Attacks %s\n", mDevice->mName->c_str(), mState.top()->entity->mName->c_str());
    }
    newState(ATTACK, mState.top()->entity);
    return;
  };

  if (inTargetRange() < 0.1)
  {
    //printf("inTargetRange() = %.2f\n", inTargetRange());
    if (mDevice->mName != NULL && mState.top()->entity->mName != NULL) {
        printf("> %s  can't Target %s\n", mDevice->mName->c_str(), mState.top()->entity->mName->c_str());
    }
    oldState(); // Back to whatever was before (possibly engage or wait).
    return;
  };

  {
      do_moveTowards();
      do_aimAt();
      do_fireAt();
  }
}


void cController::attackEnemy()
{
  if (inMeeleRange() < 0.1)
  {
    if (mDevice->mName != NULL && mState.top()->entity->mName != NULL) {
        printf("> %s can't Attack %s\n", mDevice->mName->c_str(), mState.top()->entity->mName->c_str());
    }
    oldState(); // back to target
    return;
  };

  {
      do_moveNear();
      do_aimAt();
      do_fireAt();
      //((cMech*)mDevice)->Pattern(tf, "nrnlln");
  }
}


void cController::followLeader()
{
}


void cController::fleeEnemy()
{
}


void cController::routeNav()
{
}


void cController::patrolArea()
{
}


cObject* cController::enemyNearby()
{
  return mDevice->enemyNearby();
}


float cController::inMeeleRange()
{
  return mDevice->inMeeleRange();
}


float cController::inWeaponRange()
{
  return mDevice->inWeaponRange();
}


float cController::inTargetRange()
{
  return mDevice->inTargetRange();
}


void cController::do_moveTowards()
{
  mDevice->do_moveTowards();
}


void cController::do_moveNear()
{
  mDevice->do_moveNear();
}


void cController::do_aimAt()
{
  mDevice->do_aimAt();
}


void cController::do_fireAt()
{
  mDevice->do_fireAt();
}


void cController::do_idle()
{
  mDevice->do_idle();
}

