// cController.hpp
//
// User Control and AI.
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
// aka hackCraft
//

#if !defined(CONTROLLER_INC)
#define CONTROLLER_INC

class cController;
class cObject;

#include <string>
#include <stack>
using namespace std;


class cController {
  struct state_struct {
    int state;
    string* entity_name;
    cObject* entity;
  };
  
public:
  enum { WAIT, ENGAGE, TARGET, ATTACK, FOLLOW, FLEE, ROUTE, PATROL };
  const char* getStateName(int state) {
      const char* names[] = { "WAIT", "ENGAGE", "TARGET", "ATTACK", "FOLLOW", "FLEE", "ROUTE", "PATROL" };
      return names[state];
  };
  
  public:
  // what we control
  cObject* mDevice;

public:
  stack<state_struct*> mState;
public:
  cController();
  virtual ~cController() {};

  void processSelf();

  void clearState();
  void overideState(int state, cObject* entity);
  void overideState(int state, char* entityname);
  void newState(int state, cObject* entity);
  void newState(int state, char* entityname);
  void oldState();

  // Transition Functions
  void waitEnemy();
  void engageEnemy();
  void targetEnemy();
  void attackEnemy();
  void followLeader();
  void fleeEnemy();
  void routeNav();
  void patrolArea();
  
  // Queries
  cObject* enemyNearby();
  
  // Fuzzy Predicates [0,1]
  float inMeeleRange();
  float inWeaponRange();
  float inTargetRange();
  
  // Output Actions
  void do_moveTowards();
  void do_moveNear();
  void do_aimAt();
  void do_fireAt();
  void do_idle();
  
};


#endif
