// cDome.cpp

#include "cDome.h"

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include <ctime>
#include <cstdio>
#include <cmath>

#ifndef M_PI
#define M_PI 3.14159265359
#endif


void cDome::DrawClouds(DWORD tx)
{
  glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_DEPTH_BUFFER_BIT | GL_TEXTURE_BIT);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glFogfv(GL_FOG_COLOR, mLoColor);

  // Get current Camera-Matrix.
  float m[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, m);
  // Position from 4th Col.
  float p[3];
  loop3i(p[i] = m[12+i]);
  // Reset Position on 4th Col.
  loop3i(m[12+i] = 0);
  // Apply transpose-inverse to position.
  transposeMatrix4(m);
  applyMatrix2(m, p);
  //printf("height over ground: %.2f\n", p[1]);
  //printf("Position: %.2f %.2f\n", p[0], p[2]);

  glPushMatrix();
  glLoadIdentity();
  glMultTransposeMatrixf(m); // Transpose to undo previous transpose.
  glTranslatef(0, p[Y], 0);

  clock_t c = clock();
  float e = 20.0f * M_PI / 180.0f;
  float h = 200.0f;
  float t = 0.01f * (float(c) / CLOCKS_PER_SEC);
  t = fmod(t, 1);

  glColor3fv(mLoColor);
  glBindTexture(GL_TEXTURE_2D, tx);
  float dx = 1.0f * t;
  float dz = 1.0f * t;
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, h, 0.0);
  for (int i=0; i<=18; i++) {
    glTexCoord2f(cos(i*e)*5.0+dx, sin(i*e)*5.0+dz);
    glVertex3f(cos(i*e)*2000.0, h, sin(i*e)*2000.0);
  };
  glEnd();
  glEnable(GL_DEPTH_TEST);

  glPopMatrix();
  glPopAttrib();
}


void cDome::DrawGround(DWORD tx)
{
  glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_DEPTH_BUFFER_BIT | GL_TEXTURE_BIT);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);

  // Get current Camera-Matrix.
  float m[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, m);
  // Position from 4th Col.
  float p[3];
  loop3i(p[i] = m[12+i]);
  // Reset Position on 4th Col.
  loop3i(m[12+i] = 0);
  // Apply transpose-inverse to position.
  transposeMatrix4(m);
  applyMatrix2(m, p);
  //printf("height over ground: %.2f\n", p[1]);
  //printf("Position: %.2f %.2f\n", p[0], p[2]);

  glPushMatrix();
  glLoadIdentity();
  glMultTransposeMatrixf(m); // Transpose to undo previous transpose.
  glTranslatef(0, p[Y], 0);

  float dx = -p[0]/5.0;
  float dz = -p[2]/5.0;
  glColor3fv(mLoColor);
  glBindTexture(GL_TEXTURE_2D, tx);
  
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0, 1, 0);
  glTexCoord2f(dx, dz);
  glVertex3f(0.0, 0.0, 0.0);
  float e = 20.0*M_PI/180.0f;
  for (int i=0; i<=18; i++) {
    glNormal3f(0, 1, 0);
    glTexCoord2f(cos(-i*e)*500.0+dx, sin(-i*e)*500.0+dz);
    glVertex3f(cos(-i*e)*2500.0, 0, sin(-i*e)*2500.0);
  };
  glEnd();

  glPopMatrix();
  glPopAttrib();
}


void cDome::DrawDome()
{
  glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_DEPTH_BUFFER_BIT | GL_TEXTURE_BIT);
  glDisable(GL_LIGHTING);
  glDisable(GL_FOG);
  glDisable(GL_TEXTURE_2D);
    
  // Load current Camera Matrix and
  // set Position (4th Column) = 0.
  float m[16];
  glGetFloatv(GL_MODELVIEW_MATRIX, m);
  loop3i(m[12+i] = 0);
 
  glPushMatrix();
  glLoadIdentity();
  glMultMatrixf(m);

  int steps = 8;
  float y1, y2, s1, s2, f = 90.0/steps*0.017454;
  
  float color1[3];
  memcpy(color1, mUpColor, sizeof(float)*3);
  
  float color2[3];
  memcpy(color2, mUpColor, sizeof(float)*3);
  
  float slope[3];
  set3i(slope, (mLoColor[i]-mUpColor[i])/(float)steps);

  glDisable(GL_DEPTH_TEST);
  s1 = sin(0*f);
  y1 = cos(0*f)*100;
  for (int i=1; i<=steps; i++) {
    s2 = sin(i*f);
    y2 = cos(i*f)*100;
    vectoradd(color2, color2, slope);
    glBegin(GL_TRIANGLE_STRIP);
    for (int j=0; j<=18; j++) {
      glColor3fv(color1);
      glNormal3f(s1*cos(j*20*0.017454), y1*0.01f, s1*sin(j*20*0.017454));
      glVertex3f(s1*cos(j*20*0.017454)*100, y1, s1*sin(j*20*0.017454)*100);
      glColor3fv(color2);
      glNormal3f(s2*cos(j*20*0.017454), y2*0.01f, s2*sin(j*20*0.017454));
      glVertex3f(s2*cos(j*20*0.017454)*100, y2, s2*sin(j*20*0.017454)*100);
    };
    glEnd();
    memcpy(color1, color2, sizeof(float)*3);
    y1 = y2;
    s1 = s2;
  };

  glPopMatrix();
  glPopAttrib();
}


void cDome::SetDayTime(SCALAR hour)
{
  SCALAR t = ((hour * 360.0 / 24.0) - 90.0);
  SCALAR s = sin(t*0.017454)*0.5+0.5;
  //mUpColor.Set(0.0, 0.0, s*0.8+0.2, 1.0);
  vectorset(mUpColor, 0.0, 0.0, s*0.8+0.2)
  //mLoColor.Set(s*0.95+0.05, s*0.95+0.05, s*0.95+0.05, 1.0);
  vectorset(mLoColor, s*0.95+0.05, s*0.95+0.05, s*0.95+0.05);
  mT = hour;
}


void cDome::SetSystemDayTime()
{
  time_t t = time(NULL);
  tm *l = localtime(&t);
  SCALAR lt = (l->tm_hour + l->tm_min / 60.0 + l->tm_sec / 3600.0);
  SetDayTime(lt);

//  SetDayTime((l->tm_sec)*24.0/60.0);
// ^- 60 seconds equals 24 hours - good for checking day-colors
}



