// cObject.hpp

#ifndef OBJECT_INC
#define OBJECT_INC

class cObject;
class cPad;
class cController;
class cEntityModel;

#include "cWorld.h"
#include <string>

using namespace std;


class cObject {
public:
  friend class cWorld;
  friend class cController;
  
  enum Roles {
    // Roles indicating Attributes
    NAMEABLE,
    SOCIAL,
    TRACKABLE,
    DAMAGEABLE,
    ENTITY,
    
    // Roles indicating Kind of Objects
    MECH,
    WEAPON,
    GEOSPOT,
    
    // Roles indicating Social-Groups
    RED,
    BLUE,
    GREEN,
    YELLOW,
    ARMY,
    MERCENARY,
    CIVIL,

    // Roles indicating State
    RELOADING,

    HEALTHY,
    WOUNDED,
    DEAD,

    // Special Roles
    HUMANPLAYER,
  };
  
  // BASE
  cWorld* mWorld; // The world this object was spawned in.
  bool mFragged; // Is this object scheduled to be deleted after the next frame.
  unsigned long mRoles; // Bitmask, lists the roles this object plays.

  // NAMEABLE
  string* mName;
  string* mDescription;
  unsigned int mDesignation; // { 0, .., 25 } for { a, ..., z }.

  // SOCIAL
  unsigned long mAllies; // future: Bitmask, lists Roles which Allies have.
  unsigned long mEnemies; // future: Bitmask, lists Roles which Enemies have.
  
  // TRACKABLE
  float mPos[3];
  float mRot[3];
  float mRadius;
  
  enum Parts {
      LEFT,
      RIGHT,
      BODY,
      LEGS,
      MAX_PARTS
  };
  
  // DAMAGEABLE
  int* mHp;
  //int* mSplasharmor;
  //int* mBeamarmor;
  //int* mPiercearmor;
  // int mHeat;
  // int mHeatsinks;
  // int mSplasharmor;
  // int mBeamarmor;
  // int mPiercearmor;

  // ENTITY
  cObject* mTarget;
  cPad* mPad; // Direct Human and Remote Pad-Control.
  cController* mController; // Command Stack takes controll when there is no direct Pad-Control.
  cEntityModel* mModel;
  
  // WEAPON

public:
  cObject();
  cObject(cObject* source);
  virtual ~cObject() {};

  virtual void onSpawn() {};
  virtual void onFrag() {};
  
  virtual void animateSelf(float spf) {};
  virtual void transformSelf() {};
  virtual void drawSelf() {};
  virtual void drawEffect() {};
  virtual void drawHUD() {};
  
  virtual void collideLine(float* a, float* b, cModelCollision* &result) {};
  
  // Related to States:
  
  // Queries
  virtual cObject* enemyNearby() { return NULL; };
  
  // Fuzzy Predicates [0,1]
  virtual float inMeeleRange() { return 0.0f; };
  virtual float inWeaponRange() { return 0.0f; };
  virtual float inTargetRange() { return 0.0f; };

  // Output Actions for Entities
  virtual void do_moveTowards() {};
  virtual void do_moveNear() {};
  virtual void do_aimAt() {};
  virtual void do_fireAt() {};
  virtual void do_idle() {};
};

#endif
