#include "cWeapon.h"
#include "models.h"

#include <GL/gl.h>


static void drawBall(float radius) {
    glBegin(GL_TRIANGLE_FAN);
    int n = 9;
    float step = 2*M_PI / (float) n;
    float a = 0;
    for (int i = 0; i < n; i++) {
        glVertex3f(sin(a)*radius, cos(a)*radius, 0);
        a += step;
    }
    glEnd();
}


//--------------------------------


cWeapon::cWeapon()
{
    //printf("cWeapon::cWeapon()\n");
    mMountfv = mBasefv = NULL;
    mWorld = NULL;
    mReloading = 0.0f;
}

/*
cWeapon::cWeapon(cWeapon *source)
{
    //printf("cWeapon::cWeapon(cObject*)\n");
    mBasefv    = source->mBasefv;
    mMountfv   = source->mMountfv;
    mWorld     = source->mWorld;
    mReloading = source->mReloading;
}
*/

//--------------------------------


void cMachineGun::fireSelf(cObject *target)
{
    if (mReloading > 0) return;
    mReloading = 0.4;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    Shrapnel* s = new Shrapnel();
    s->cnt = 4;
    
    s->pos[0] = source[12];
    s->pos[1] = source[13];
    s->pos[2] = source[14];

    float nrm[3];
    float pos2[] = { 0, 0, 1 };
    applyMatrix2(source, pos2);
    vectorsub(nrm, pos2, s->pos);
    
    scalarmul(s->vel, nrm, 200);
    const float variation = 0.02;
    for (int i=0; i<3; i++) s->vel[i] += (50-(rand()%100)) * variation;
    shrapnels.push_back(s);
    
    float len1 = 1;
    float len2 = 30;
    float a[] = {
        s->pos[0]+nrm[0]*len1,
        s->pos[1]+nrm[1]*len1,
        s->pos[2]+nrm[2]*len1
    };
    float b[] = {
        s->pos[0]+nrm[0]*len2,
        s->pos[1]+nrm[1]*len2,
        s->pos[2]+nrm[2]*len2
    };
    
    delete ( mWorld->filterByCollideLine(NULL, a, b, -1) );
}


void cMachineGun::animateSelf(float spf)
{
    for (list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            shrapnels.remove(s);
            delete s;
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cMachineGun::drawEffect()
{
    glPushMatrix();
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDepthMask(GL_FALSE);

    for (list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        const float len1 = 0.01;
        const float len2 = 0.02;
        glBegin(GL_LINE_STRIP);
            glColor4f(1,1,1,1);
            glVertex3fv(s->pos);
            glColor4f(1,1,0,1);
            glVertex3f(s->pos[0] - s->vel[0]*len1, s->pos[1] - s->vel[1]*len1, s->pos[2] - s->vel[2]*len1);
            glColor4f(1,0,0,0.3);
            glVertex3f(s->pos[0] - s->vel[0]*len2, s->pos[1] - s->vel[1]*len2, s->pos[2] - s->vel[2]*len2);
        glEnd();
    }

    glPopMatrix();
    glPopAttrib();
}


//--------------------------------


void cPlasmaGun::fireSelf(cObject *target)
{
    if (mReloading > 0) return;
    mReloading = 0.4;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    Shrapnel* s = new Shrapnel();
    s->cnt = 4;
    
    s->pos[0] = source[12];
    s->pos[1] = source[13];
    s->pos[2] = source[14];

    float nrm[3];
    float pos2[] = { 0, 0, 1 };
    applyMatrix2(source, pos2);
    vectorsub(nrm, pos2, s->pos);
    
    scalarmul(s->vel, nrm, 100);
    shrapnels.push_back(s);
    
    float len1 = 1;
    float len2 = 30;
    float a[] = {
        s->pos[0]+nrm[0]*len1,
        s->pos[1]+nrm[1]*len1,
        s->pos[2]+nrm[2]*len1
    };
    float b[] = {
        s->pos[0]+nrm[0]*len2,
        s->pos[1]+nrm[1]*len2,
        s->pos[2]+nrm[2]*len2
    };
    
    delete ( mWorld->filterByCollideLine(NULL, a, b, -1) );
}


void cPlasmaGun::animateSelf(float spf)
{
    for (list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            shrapnels.remove(s);
            delete s;
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cPlasmaGun::drawEffect()
{
    glPushMatrix();
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDepthMask(GL_FALSE);

    float t[3];
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);
    memcpy(t, &m[12], sizeof(float)*3);
    m[12] = m[13] = m[14] = 0;
    float n[] = { // = m Transposed = m Inverted
        m[0], m[4], m[ 8], 0,
        m[1], m[5], m[ 9], 0,
        m[2], m[6], m[10], 0,
           0,    0,     0, 1
    };

    for (list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                glColor4f(0.1, 0.1, 0.6, 0.99f);
                drawBall(1*0.1f);
                glColor4f(0.8, 0.8, 0.1, 0.6f);
                drawBall(1*0.07f);
        glPopMatrix();
    }

    glPopMatrix();
    glPopAttrib();
}


//--------------------------------


void cSparkGun::fireSelf(cObject *target)
{
    if (mReloading > 0) return;
    mReloading = 0.8;
    
    float source[16];
    glPushMatrix();
        glLoadIdentity();
            glMultMatrixf(mBasefv);
                glMultMatrixf(mMountfv);
                glGetFloatv(GL_MODELVIEW_MATRIX, source);
    glPopMatrix();
    
    Shrapnel* s = new Shrapnel();
    s->cnt = 4;
    
    s->pos[0] = source[12];
    s->pos[1] = source[13];
    s->pos[2] = source[14];

    float nrm[3];
    float pos2[] = { 0, 0, 1 };
    applyMatrix2(source, pos2);
    vectorsub(nrm, pos2, s->pos);
    
    scalarmul(s->vel, nrm, 60);
    shrapnels.push_back(s);
    
    float len1 = 1;
    float len2 = 30;
    float a[] = {
        s->pos[0]+nrm[0]*len1,
        s->pos[1]+nrm[1]*len1,
        s->pos[2]+nrm[2]*len1
    };
    float b[] = {
        s->pos[0]+nrm[0]*len2,
        s->pos[1]+nrm[1]*len2,
        s->pos[2]+nrm[2]*len2
    };
    
    delete ( mWorld->filterByCollideLine(NULL, a, b, -1) );
}


void cSparkGun::animateSelf(float spf)
{
    for (list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        s->pos[0] += s->vel[0] * spf;
        s->pos[1] += s->vel[1] * spf;
        s->pos[2] += s->vel[2] * spf;
        s->cnt -= spf;
        if (s->cnt < 0) {
            shrapnels.remove(s);
            delete s;
        }
    }
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
};


void cSparkGun::drawEffect()
{
    glPushMatrix();
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_ALL_ATTRIB_BITS);

    glDisable(GL_CULL_FACE);
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDepthMask(GL_FALSE);

    float t[3];
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);
    memcpy(t, &m[12], sizeof(float)*3);
    m[12] = m[13] = m[14] = 0;
    float n[] = { // = m Transposed = m Inverted
        m[0], m[4], m[ 8], 0,
        m[1], m[5], m[ 9], 0,
        m[2], m[6], m[10], 0,
           0,    0,     0, 1
    };

    for (list<Shrapnel*>::iterator i = shrapnels.begin(); i != shrapnels.end();) {
        Shrapnel* s = *i++;
        glPushMatrix();
            glTranslatef(s->pos[0], s->pos[1], s->pos[2]);
                glMultMatrixf(n);
                glColor4f(0.6, 0.1, 0.1, 0.99f);
                drawBall(5*0.1f);
                glColor4f(0.8, 0.8, 0.1, 0.3f);
                drawBall(5*0.07f);
        glPopMatrix();
    }

    glPopMatrix();
    glPopAttrib();
}


//--------------------------------


void cRaybeam::fireSelf(cObject *target)
{
    if (mReloading > 0 || raylife > 0) return;
    mReloading = 1.0;
    raylife = 0.7;
    //mReloading = raylife;
}


void cRaybeam::animateSelf(float spf)
{
    raylife -= spf;
    if (raylife < 0) raylife = 0.0f;
    mReloading -= spf;
    if (mReloading < 0) mReloading = 0.0f;
}


void cRaybeam::drawSelf()
{
};


void cRaybeam::drawEffect()
{
    if (raylife == 0) return;
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glDepthMask(GL_FALSE);
        
        glPushMatrix();
            glMultMatrixf(this->mBasefv);
            glMultMatrixf(this->mMountfv);

            float m[16];
            glGetFloatv(GL_MODELVIEW_MATRIX, m);
            float* mpnt = &m[12]; // Vector that points to the Mountpoint relative to the eye.
            //printVector(mpnt);
            float* mfwd = &m[8]; // Mountpoint-Forward Vector relative to the eye.
            //printVector(mfwd);
            
            float right[3];
            cross(right, mpnt, mfwd);
            normalize(right, right);
            //printVector(right);
            
            if (mpnt[2] > 0 || mfwd[2] < 0) right[1] = -right[1]; // Correction
            //if (mpnt[1] > 0 || mfwd[1] < 0) right[1] = -right[1]; // Correction
            
            const float len = 10;
            struct appearance {
                float width, length;
                float icolor[4];
                float ocolor[4];
            } app[] = {
                0.10,   8*len,  {0.3, 0.3, 0.3, 1},  {0.3, 0.0, 0.0, 1},
                0.40,   8*len,  {0.5, 0.5, 0.5, 1},  {0.3, 0.0, 0.0, 1},
                0.80,   8*len,  {0.9, 0.9, 0.9, 1},  {0.3, 0.0, 0.0, 1},
                0.85,   8*len,  {1.0, 1.0, 1.0, 1},  {0.4, 0.0, 0.0, 1},
                0.90, 8.2*len,  {1.0, 1.0, 1.0, 1},  {0.7, 0.0, 0.0, 1},
                0.60, 8.4*len,  {0.9, 0.9, 0.9, 1},  {0.3, 0.0, 0.0, 1},
                0.30, 8.5*len,  {0.7, 0.7, 0.7, 1},  {0.3, 0.0, 0.0, 1},
            };
            
            float width[] = { 0.1, 0.4, 0.8, 0.85, 0.9, 0.6, 0.3 };
            
            int lifeindex = (int) ((0.7-raylife) * 10);
            if (lifeindex > sizeof(width)/sizeof(float) - 1) lifeindex = sizeof(width)/sizeof(float) - 1;
            if (lifeindex < 0) lifeindex = 0;
            const float w = app[lifeindex].width;
            
            right[0] *= w;
            right[1] *= w;
            right[2] *= w;
            glBegin(GL_TRIANGLE_STRIP);
                glColor4fv(app[lifeindex].ocolor);
                glVertex3f(-right[0], -right[1], -right[2]+2*w); glVertex3f(-right[0], -right[1], -right[2]+app[lifeindex].length);
                glColor4fv(app[lifeindex].icolor);
                glVertex3f(0, 0.0, 0); glVertex3f(0, 0.0, +app[lifeindex].length);
                glColor4fv(app[lifeindex].ocolor);
                glVertex3f(+right[0], +right[1], +right[2]+2*w); glVertex3f(+right[0], +right[1], +right[2]+app[lifeindex].length);
            glEnd();
            
        glPopMatrix();
    glPopAttrib();
};


//--------------------------------

