// cWorld.hpp

#if !defined(WORLD_INC)
#define WORLD_INC

class cWorld;

#include "psi3d/psi3d.h"
#include <list>

using namespace std;

class cDome;
class cObject;
class cEntityModel;
class cModel;

struct cModelCollision {
  cEntityModel* intsec_entmod;
  float intsec_distance;
  float intsec_point[3];
  float intsec_normal[3];
};

#define loopObjects(objset)   for (list<cObject *>::iterator it = (objset).begin(); it != (objset).end();)

class cWorld {
    static cWorld* sWorld;

public: // Deprecated, need to change:

  cDome* mDome; // Going to be a separated cObject.
  float mGravity[3]; // Downward force vector (or 0,0,0 in space).

protected: // Attributes:

  list<cObject*> mObjects; // Contains *every* object which isn't yet completly deleted.
  list<cObject*> mCorpses[2]; // Doublebuffer for object delete scheduling.
  list<cObject*>* mBufferedCorpses; // Points to current one of the two delete doublebuffers.

  cModel* mParticleModel;
  cModel* mPlantModel;
  cModel* mConstructModel;

  //map<string, void *> mGamevars;
  //map<string, string> mGamevars;

  float mTime; // Seconds since simulation begin.

  unsigned long mOldClock; // Previous frame's time (for spf calculation).
  unsigned long mNewClock; // Current frame's time (for spf calculation).

  float mSPF; // seconds per frame (1.0f/fps).
  float mFPS; // frames per second (1.0f/spf).

public: // Constructor, Object-Management and Drawing:

  cWorld();
  
public: // Spawning And Fragging

  // Spawns object imediatly.
  void spawnObject(cObject *object);
  
  // Sets object fragged and inserts it into the mBufferedCorpses-List
  // so it can be deleted later on by bagFragged().
  void fragObject(cObject *object);
  
  // Call after each frame: Deletes fragged objects of previous frame.
  // Fragged Objects are only removed from the mObjects-List here.
  void bagFragged();

public: // Get/Set:

  float getSPF() { return mSPF; };
  float getFPS() { return mFPS; };
  float getTime() { return mTime; };

  float getGravity(int coord) { return mGravity[coord]; };
  
  cModel* getParticleModel() { return mParticleModel; }
  cModel* getPlantModel() { return mPlantModel; }
  cModel* getConstructModel() { return mConstructModel; }

public: // Simulation Step.

  // Call every frame to update (animation) timers.
  void updateTimers(); 

  void animateSelf(float spf);
  void transformSelf();
  void drawSelf();
  void drawEffect();

public: // World-Scanning, World-Sense for objects.
  
  // Debug function to print out an object-list.
  // With a parameter of NULL/no parameter at all prints all objects within the world.
  // Warning: Printing all objects could lead to massive load on the console. 
  void printNames(list<cObject*>* objects = NULL);

  // Filters the given objectlist (or all mObjects if NULL) by roles (ORed Bitmask) (excluding ex Object).
  // There is the option to match all roles (true=AND) or to select any given role (false=OR).
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByRole(cObject* ex, unsigned long rolemask, bool all, int maxamount, list<cObject*>* objects = NULL);
  
  // Returns a List of objects which are within minimum and maximum range (excluding ex Object).
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByRange(cObject* ex, float* origin, float minrange, float maxrange, int maxamount, list<cObject*>* objects = NULL);
  
  // Returns a List of objects which have the specified name.
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByName(cObject* ex, char* name, int maxamount, list<cObject*>* objectlist = NULL);
    
  // Returns a List of objects which have intersections with a line (excluding ex Object).
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByCollideLine(cObject* ex, float* pointa, float* pointb, int maxamount, list<cObject*>* objects = NULL);

  // Returns a List of objects whose bounding volume (approximately) intersects with the cylinder
  // described by the volume of certain diameter around a line.
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByCylinder(cObject* ex, float* pointa, float* pointb, float diameter, int maxamount, list<cObject*>* objects = NULL);

public: // Misc

  static const char* getNATO(int letter);
};

#endif
