// entities.hpp

#if !defined(ENTITIES_INC)
#define ENTITIES_INC

class cGeospot;
class cWeapon;
class cBuilding;
class cTree;
class cMech;
class cTank;

#include "cObject.h"

class cWeapon;

//-----------------------------------


class cGeospot : public cObject {
  public:
  cGeospot(float* pos = NULL, float* rot = NULL, char* domain_name = NULL, unsigned int des = 0);
};


class cBuilding : public cObject {
  public:
  cBuilding(float* pos = NULL, float* rot = NULL);

  virtual void onSpawn();

  virtual void transformSelf();
  virtual void drawSelf();
};


class cTree : public cObject {
  protected:
  string desc;
  public:
  cTree(float* pos = NULL, float* rot = NULL, char* descriptor = NULL);

  virtual void onSpawn();

  virtual void transformSelf();
  virtual void drawSelf();
};


class cMech : public cObject {
  public:
  enum {
    Zero,Hips, CTorsor, LTorsor, RTorsor, Head,
    LUpLeg, RUpLeg, LLoLeg, RLoLeg, LFoot, RFoot,
    LUpArm, RUpArm, LLoArm, RLoArm, MAX_MECH_PARTS
  };

  protected:
  float mTimer;
  
  list<cWeapon *> mWeapons;
  int mCurrentWeapon;
  
  float mTwr[3];
  float mJetOn;
  float mSpeed;

  float mMatrix[16];
  float (*mRotators)[3];
  float mPilot[16];

  public:
  cMech(float* pos = NULL, float* rot = NULL);

  virtual void onSpawn();

  float* getPilotMatrix();
  void Pattern(float t, char *pattern);
  void Walk(float spf);

  void ChassisLR(float r);
  void ChassisUD(float r);
  void TowerLR(float r);
  void TowerUD(float r);

  void fireAllWeapons(cObject *target);
  void fireWeapon(int n, cObject *target);
  void mountWeapon(char* point, cWeapon *weapon);

  virtual void animateSelf(float spf);
  virtual void transformSelf();
  virtual void drawSelf();
  virtual void drawEffect();
  virtual void drawHUD();

  virtual void collideLine(float* a, float* b, cModelCollision* &result);
  
  // Queries
  virtual cObject* enemyNearby();
  
  // Fuzzy Predicates [0,1]
  virtual float inMeeleRange();
  virtual float inWeaponRange();
  virtual float inTargetRange();
  
  // Output Actions
  virtual void do_moveTowards();
  virtual void do_moveNear();
  virtual void do_aimAt();
  virtual void do_fireAt();
  virtual void do_idle();
};

/*
class cTank : public cEntity {
  public:
  enum { Zero,Chassis, LChain, RChain, Turret, Weapon };
  public:
};
*/

#endif
