// main.cpp
//
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
// aka hackCraft
//

#include "main.h"
#include <GL/glut.h>

FILE* errlog = NULL; // Legacy Dependency from old psi3d

cPad                  gPad1, gPad2; // Player One (Pilot?) and Two (Gunner?) Gamepads.
cPad*                 gKeyPad = &gPad1; // Keyboard inputs send to which Gamepad?

cObject*              gCameraEntity = NULL;

cConio3d             *con[2], *gCurrentConsole;

cViewer               gViewer;

cWorld               *planet;

int                   width       = 0;
int                   height      = 0;


bool                  external_camera = false;
bool                  satelite_camera = false;
bool                  paused      = false;
bool                  fullscreen  = false;
bool                  wireframe   = false;

WORD                  TexID[100];


struct Mouse {
    Mouse() { memset(this, 0, sizeof(Mouse)); }
    unsigned short buttons;
    float speed[2];
    float pos[2];
} gMouse;


void setupMatrix(float *mat, float *rot, float *pos)
{
  glPushMatrix();
  glLoadIdentity();

  if (pos != NULL) glTranslatef(pos[0], pos[1], pos[2]);

  if (rot != NULL) {
    glRotatef(rot[0], 1, 0, 0);
    glRotatef(rot[1], 0, 1, 0);
    glRotatef(rot[2], 0, 0, 1);
  };

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}


void setupInvMatrix(float *mat, float *rot, float *pos)
{
  glPushMatrix();
  glLoadIdentity();

  if (rot != NULL) {
    glRotatef(-rot[2], 0, 0, 1);
    glRotatef(-rot[1], 0, 1, 0);
    glRotatef(-rot[0], 1, 0, 0);
  };
  if (pos != NULL) glTranslatef(-pos[0], -pos[1], -pos[2]);

  glGetFloatv(GL_MODELVIEW_MATRIX, mat);

  glPopMatrix();
}



void Fatal(char *errorString)
{
  fprintf(stderr, "%s\n", errorString);
  fflush(stderr);
  exit(1);
}


BOOLEAN LoadTgaTex(char *fname, BOOLEAN transparency, BOOLEAN mipmap, WORD &TexID)
{
  cBmp *b = new cBmp;
  cVector pink;

  if (b->LoadTga(fname)) return 1;
  if (transparency) {
    pink.Set(255, 0, 255);
    b->MakeAlpha(pink);
    b->Mul(W, 0.7f); // make more transparent
  };
  return b->glRegister(mipmap, TexID);
  delete [] b;
}


void initHUD() {
  con[0] = new cConio3d;
  con[0]->Resize(16, 8, 0.02);
  con[0]->SetFont(TexID[4]);
  con[0]->Insert("--------------\n");  
  con[0]->Insert("Lin Warrior 3D\n");  
  con[0]->Insert("--------------\n");  
  con[0]->Insert("              \n");  
  con[0]->Insert("3D Console 0  \n");  
  con[0]->Insert("              \n");  
  con[0]->Insert("Status: INACTIVE\n");  
  con[0]->mPose.mPos.mV[X] = +0.29;
  con[0]->mPose.mPos.mV[Y] = +0.27;
  con[0]->mPose.mRot.mV[X] = +25;
  con[0]->mPose.mRot.mV[Y] = -25;
  con[0]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  con[1] = new cConio3d;
  con[1]->Resize(16, 8, 0.02);
  con[1]->SetFont(TexID[4]);
  con[1]->Insert("--------------\n");  
  con[1]->Insert("Lin Warrior 3D\n");  
  con[1]->Insert("--------------\n");  
  con[1]->Insert("              \n");  
  con[1]->Insert("3D Console 1  \n");  
  con[1]->Insert("              \n");  
  con[1]->Insert("Status: ACTIVE\n");  
  con[1]->mPose.mPos.mV[X] = -0.29;
  con[1]->mPose.mPos.mV[Y] = +0.27;
  con[1]->mPose.mRot.mV[X] = +25;
  con[1]->mPose.mRot.mV[Y] = +25;
  con[1]->mCenter.Set(-8.0*0.02, 4*0.02, 0);

  gCurrentConsole = con[1];
}


void initMechs() {
  if (true) {
    float pos[3];
    pos[0] = -50;
    pos[1] = 0;
    pos[2] = -50;
    cMech* mech = new cMech(pos);
    mech->mName = new string("Alpha");
    assert(mech != NULL);

    gCameraEntity = mech;
    mech->mPad = &gPad1;

    planet->spawnObject(mech);

    if (true) {
      mech->mountWeapon("LTorsor", new cPlasmaGun);
      mech->mountWeapon("RTorsor", new cSparkGun);
      mech->mountWeapon("RLoArm",  new cMachineGun);
      mech->mountWeapon("LLoArm",  new cRaybeam);
    }
  }


  if (true) {
    float pos[3];
    pos[0] = 50;
    pos[1] = 0;
    pos[2] = -50;
    cMech* mech = new cMech(pos);
    mech->mName = new string("Bravo");
    assert(mech != NULL);

    cController* ai = new cController;
    ai->mDevice = mech;
    mech->mController = ai;
    mech->mPad = new cPad();
    //ai->newState(cControllerI::TARGET, "Charly");
    //ai->newState(cController::ROUTE, "Attackroute");

    planet->spawnObject(mech);

    if (true) {
      mech->mountWeapon("LLoArm",  new cMachineGun);
      mech->mountWeapon("RLoArm",  new cMachineGun);
      mech->mountWeapon("RTorsor",  new cMachineGun);
      mech->mountWeapon("LTorsor",  new cMachineGun);
    }
  }


  if (true) {
    float pos[3];
    pos[0] = 50;
    pos[1] = 0;
    pos[2] = 50;
    cMech* mech = new cMech(pos);
    mech->mName = new string("Charly");
    assert(mech != NULL);

    cController* ai = new cController;
    ai->mDevice = mech;
    mech->mController = ai;
    mech->mPad = new cPad();
    //ai->newState(cController::WAIT, (char*) NULL);
    //ai->newState(cController::TARGET, "Delta");
    //ai->newState(cController::ROUTE, "Attackroute");

    planet->spawnObject(mech);

    if (true) {
      mech->mountWeapon("LTorsor",  new cPlasmaGun);
      mech->mountWeapon("RTorsor",  new cSparkGun);
    }
  }


  if (true) {
    float pos[3];
    pos[0] = -50;
    pos[1] = 0;
    pos[2] = 50;
    cMech* mech = new cMech(pos);
    mech->mName = new string("Delta");
    assert(mech != NULL);

    cController* ai = new cController;
    ai->mDevice = mech;
    mech->mController = ai;
    mech->mPad = new cPad();
    ai->newState(cController::ENGAGE, "Charly");
  //  ai->newState(cController::ROUTE, navs);

    planet->spawnObject(mech);

    if (true) {
      mech->mountWeapon("LTorsor",  new cSparkGun);
      mech->mountWeapon("RTorsor",  new cRaybeam);
    }
  }
}


void initArchitecture() {
  if (!true) {
    int bld_dist = 50;
    cBuilding* bld;
    for (int i=0; i<3; i++) {
      for (int j=0; j<3; j++) {
        if (!((i==1) && (j==1))) {
          float pos[3];
          pos[0] = i*bld_dist*2-bld_dist;
          pos[1] = 0;
          pos[2] = j*bld_dist*2-bld_dist;
          bld = new cBuilding(pos);
          planet->spawnObject(bld);
        };
      };
    };
  }
}


void initFlora() {
  if (true) {
    int dist = 320;
    float a, b;
    float pos[3];
    float rot[3];
    cTree* tr;

    for (int i=0; i<200; i++) {
      a = rand()%360;
      //b = dist + rand() % (2*dist);
      b = rand() % (2*dist);
      pos[0] = sin(a*0.017454)*b;
      pos[1] = 0;
      pos[2] = cos(a*0.017454)*b;
      rot[0] = 0;//10-rand()%20;
      rot[1] = 10*(rand()%36);
      rot[2] = 0;//10-rand()%20;
      tr = new cTree(pos, rot, "3 5.7");
      planet->spawnObject(tr);
    };

    for (int i=0; i<9; i++) {
      a = dist - rand() % (2*dist);
      b = dist - rand() % (2*dist);
      pos[0] = a;
      pos[1] = 0;
      pos[2] = b;
      rot[0] = 0;//10-rand()%20;
      rot[1] = 10*(rand()%36);
      rot[2] = 0;//10-rand()%20;
      tr = new cTree(pos, rot, "2 2.0");
      planet->spawnObject(tr);
    };

    for (int i=0; i<6; i++) {
      a = dist - rand() % (2*dist);
      b = dist - rand() % (2*dist);
      pos[0] = a;
      pos[1] = 0;
      pos[2] = b;
      rot[0] = 0;//10-rand()%20;
      rot[1] = 10*(rand()%36);
      rot[2] = 0;//10-rand()%20;
      tr = new cTree(pos, rot, "3 2.0");
      planet->spawnObject(tr);
    };

    for (int i=0; i<3; i++) {
      a = dist - rand() % (2*dist);
      b = dist - rand() % (2*dist);
      pos[0] = a;
      pos[1] = 0;
      pos[2] = b;
      rot[0] = 0;//10-rand()%20;
      rot[1] = 10*(rand()%36);
      rot[2] = 0;//10-rand()%20;
      tr = new cTree(pos, rot, "4 2.0");
      planet->spawnObject(tr);
    };

    for (int i=0; i<0; i++) {
      a = dist - rand() % (2*dist);
      b = dist - rand() % (2*dist);
      pos[0] = a;
      pos[1] = 0;
      pos[2] = b;
      rot[0] = 0;//10-rand()%20;
      rot[1] = 10*(rand()%36);
      rot[2] = 0;//10-rand()%20;
      tr = new cTree(pos, rot, "5 2.0");
      planet->spawnObject(tr);
    };
  }
}


void initMission() {
  planet = new cWorld;
  planet->mDome = new cDome;

  initMechs();
  initArchitecture();
  initFlora();
}


void Init()
{
  //float Black[3] = { 0.0, 0.0, 0.1 };
  //float lightColor[3] = { 0.5, 0.2, 1.0 };
  float fogColor[4] = { 1.0, 1.0, 1.0, 1.0 };
  float fogDensity = 0.001;

//  glDisable(GL_DITHER);
  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);

  glClearColor(0.0, 0.125, 0.0, 0.0);
  glClearDepth(1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//  glShadeModel(GL_FLAT);
  glShadeModel(GL_SMOOTH);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthMask(GL_TRUE);
  glDepthFunc(GL_LEQUAL);

  glFogi(GL_FOG_MODE, GL_EXP);
  glFogfv(GL_FOG_COLOR, fogColor);
  glFogf(GL_FOG_DENSITY, fogDensity);

//  glHint(GL_FOG_HINT,GL_FASTEST);

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
//  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

#define LIGHTON
#ifdef LIGHTON
  glEnable(GL_LIGHTING);
  float p[] = { 0, 100, 0, 1 };
  glLightfv(GL_LIGHT0, GL_POSITION, p);
  glEnable(GL_COLOR_MATERIAL);
#endif

  if (
    //LoadTgaTex("data/textures/house0.tga",    0,1, TexID[0]) |
    LoadTgaTex("data/textures/clouds0.tga",   1,1, TexID[3]) |
    LoadTgaTex("data/textures/font0.tga",     1,0, TexID[4]) |
    //LoadTgaTex("data/textures/flare0.tga",    1,1, TexID[5]) |
    LoadTgaTex("data/textures/tile0.tga",     0,1, TexID[6]) |
    LoadTgaTex("data/textures/ground0.tga",   1,1, TexID[7])
    //| LoadTgaTex("data/textures/tree0.tga",     1,1, TexID[8]) 
  ) Fatal("could not load tga texture");

  initHUD();

  initMission();
}


void Reshape(int w, int h)
{
  width = w;
  height = h;
  gViewer.mType = 0;
  gViewer.Move(0, 0);
  gViewer.Resize(width, height);
  glutPostRedisplay();
}


void Draw()
{
  if (paused) return;

  static float tf = 0;


// Setup Planets Time (to determine SkyDome colors)
//
  
  cDome *dome = planet->mDome;

  dome->SetSystemDayTime();      // local time of "real" world
  //dome->SetDayTime(12.00);     // High Noon!!!
  //dome->SetDayTime(24.00);     // Mid Night!!!
  dome->SetDayTime(19.00);


// Setup to prepare to Draw
//


  if (!wireframe) {
    glClearColor(dome->mLoColor[X], dome->mLoColor[Y], dome->mLoColor[Z], 1.0);
    glFogfv(GL_FOG_COLOR, dome->mLoColor);
  } else {
    float dark[] = { 0.0, 0.0, 0.0, 1.0 };
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glFogfv(GL_FOG_COLOR, dark);
  };

  glEnable(GL_FOG);
#ifdef LIGHTON
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
#endif

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  gViewer.Use();


  // Do Animtations and To-World-Transformations.
  //

  planet->animateSelf(planet->getSPF());

  glPushMatrix();
    glLoadIdentity();
      planet->transformSelf();
  glPopMatrix();


  // Setup camera.
  //

  cMech* mech = (cMech*) gCameraEntity;
  
  float *matrixfv = mech->getPilotMatrix();
  assert(matrixfv != NULL);
  
  glLoadIdentity();
  if (external_camera) {
    glTranslatef(0, 0, -5);
    glRotatef(15, 1, 0, 0);
  } else if (satelite_camera) {
    glTranslatef(0, 0, -100);
    glRotatef(90, 1, 0, 0);
  };
  glRotatef(180, 0, 1, 0);
  glMultMatrixf(matrixfv);


#if 0
  // Debug: Draw Lines from Mech No.1 to Mech No.2
  //
  glDisable(GL_TEXTURE_2D);
  if (!true) {
    glColor4f(1.0, 0.0, 0.0, 1.0);
    glBegin(GL_LINES);
      glVertex3fv(theAI->mSelf.mV);
      glVertex3fv(theAI->mDest.mV);
    glEnd();
  
    glColor4f(0.0, 1.0, 0.0, 1.0);
    glBegin(GL_LINES);
      glVertex3fv(theAI2->mSelf.mV);
      glVertex3fv(theAI2->mDest.mV);
    glEnd();
  }
#endif

// Setup the drawing scheme

  if (!wireframe) {
    glEnable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glEnable(GL_CULL_FACE);
  } else {
    glDisable(GL_TEXTURE_2D);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glDisable(GL_CULL_FACE);
  };

// Draw the environment: Ground, SkyDome, Clouds
//
  dome->DrawGround(TexID[7]);

  if (!wireframe) {
    dome->DrawDome();
    dome->DrawClouds(TexID[3]);
  };


// Draw the things: Mechs, Buildings, Trees, Shots, ...
//
  
  glBindTexture(GL_TEXTURE_2D, TexID[6]);

  planet->drawSelf();
  
  glDisable(GL_CULL_FACE);

  planet->drawEffect();

// Draw the 3d Consoles
//

  if (!true) {
    cMatrix b;
    b.Set();
    b.MovRel(0, 0, -1);
  
    con[0]->mPose.mRot.mV[Y] += sin(5*tf*0.017454)*0.1;
    con[0]->Draw(b);
  
    con[1]->mPose.mRot.mV[Y] -= sin(5*tf*0.017454)*0.1;
    con[1]->Draw(b);
  }
  
  gCameraEntity->drawHUD();

// Make the drawn things visible and setup some counters
//

  glFlush();
  glutSwapBuffers();

  tf+=0.3;

  planet->bagFragged(); // Finaly delete fragged.
  planet->updateTimers(); // calculate the time that frame took
  
  //gPad1.print();
}


void Idle()
{
  glutPostRedisplay();
}


void MouseFunc(int button, int state, int x, int y)
{
  if (button == GLUT_RIGHT_BUTTON)   if (state == GLUT_DOWN) gMouse.buttons |= 2; else gMouse.buttons &= !2;
  if (button == GLUT_MIDDLE_BUTTON)  if (state == GLUT_DOWN) gMouse.buttons |= 4; else gMouse.buttons &= !4;
  if (button == GLUT_LEFT_BUTTON)    if (state == GLUT_DOWN) gMouse.buttons |= 1; else gMouse.buttons &= !1;
  if (x > width) x = width;
  if (y > height) y = height;
  gMouse.pos[0] = x;
  gMouse.pos[1] = y;
  gMouse.speed[0] = (y-height/2.)/(height/4.);
  gMouse.speed[1] = (x-width/2.)/(width/4.);
}


void Mouse(int x, int y)
{
  if (x > width) x = width;
  if (y > height) y = height;
  gMouse.pos[0] = x;
  gMouse.pos[1] = y;
  gMouse.speed[0] = (y-height/2.)/(height/4.);
  gMouse.speed[1] = (x-width/2.)/(width/4.);
}


void KeyDn(unsigned char k, int x, int y)
{
  if (k == 27) {
    exit(0);
  };

  if (k == CAMERA_KEY) {
    external_camera = !external_camera;
    satelite_camera = 0;
  } else if (k == SAT_CAMERA_KEY) {
    satelite_camera = !satelite_camera;
    external_camera = 0;
  } else if (k == WIREFRAME_KEY) {
    wireframe = !wireframe;
  } else if (k == PAUSE_KEY) {
    if (!paused) {
      gCurrentConsole->Insert("\nPAUSED\n");
      Draw();
    } else {
      gCurrentConsole->Insert("\nSUSPEND\n");
    };
    paused = !paused;
  } else {
    if (gCurrentConsole != NULL) gCurrentConsole->InsertChar(k);
  };
  
  int mod = glutGetModifiers();
  
  if (mod != GLUT_ACTIVE_ALT) {
    if (k == 'j') gKeyPad->setAxis(cPad::AX_LR2, -1.0f);
    if (k == 'l') gKeyPad->setAxis(cPad::AX_LR2, +1.0f);
    if (k == 'i') gKeyPad->setAxis(cPad::AX_UD2, -1.0f);
    if (k == 'k') gKeyPad->setAxis(cPad::AX_UD2, +1.0f);
  } else if (mod == GLUT_ACTIVE_ALT) {
    if (k == 'j') gKeyPad->setButton(cPad::BT_HL, true);
    if (k == 'l') gKeyPad->setButton(cPad::BT_HR, true);
    if (k == 'i') gKeyPad->setButton(cPad::BT_HU, true);
    if (k == 'k') gKeyPad->setButton(cPad::BT_HD, true);
  };
  
  if (k == 'a') gKeyPad->setButton(cPad::BT_L1, true);
  if (k == 'g') gKeyPad->setButton(cPad::BT_R1, true);
  
  if (k == 'w') gKeyPad->setButton(cPad::BT_L2, true);
  if (k == 'r') gKeyPad->setButton(cPad::BT_R2, true);
  
  if (k == 's') gKeyPad->setButton(cPad::BT_SQUARE, true);
  if (k == 'f') gKeyPad->setButton(cPad::BT_CIRCLE, true);
  
  if (k == 'e') gKeyPad->setButton(cPad::BT_DELTA, true);
  if (k == 'd') gKeyPad->setButton(cPad::BT_CROSS, true);
  
  if (k == '.') gKeyPad->setButton(cPad::BT_J1B, true);
  if (k == ',') gKeyPad->setButton(cPad::BT_J2B, true);
}

void KeyUp(unsigned char k, int x, int y)
{
  if (k == 'j') gKeyPad->setAxis(cPad::AX_LR2, 0.0f);
  if (k == 'l') gKeyPad->setAxis(cPad::AX_LR2, 0.0f);
  if (k == 'i') gKeyPad->setAxis(cPad::AX_UD2, 0.0f);
  if (k == 'k') gKeyPad->setAxis(cPad::AX_UD2, 0.0f);
  
  if (k == 'j') gKeyPad->setButton(cPad::BT_HL, false);
  if (k == 'l') gKeyPad->setButton(cPad::BT_HR, false);
  if (k == 'i') gKeyPad->setButton(cPad::BT_HU, false);
  if (k == 'k') gKeyPad->setButton(cPad::BT_HD, false);
  
  if (k == 'a') gKeyPad->setButton(cPad::BT_L1, false);
  if (k == 'g') gKeyPad->setButton(cPad::BT_R1, false);
  
  if (k == 'w') gKeyPad->setButton(cPad::BT_L2, false);
  if (k == 'r') gKeyPad->setButton(cPad::BT_R2, false);
  
  if (k == 's') gKeyPad->setButton(cPad::BT_SQUARE, false);
  if (k == 'f') gKeyPad->setButton(cPad::BT_CIRCLE, false);
  
  if (k == 'e') gKeyPad->setButton(cPad::BT_DELTA, false);
  if (k == 'd') gKeyPad->setButton(cPad::BT_CROSS, false);
  
  if (k == '.') gKeyPad->setButton(cPad::BT_J1B, false);
  if (k == ',') gKeyPad->setButton(cPad::BT_J2B, false);
}


void SpecialDn(int k, int x, int y)
{
  if (k == GLUT_KEY_LEFT) gKeyPad->setAxis(cPad::AX_LR1, -1.0f);
  if (k == GLUT_KEY_RIGHT) gKeyPad->setAxis(cPad::AX_LR1, +1.0f);
  if (k == GLUT_KEY_UP) gKeyPad->setAxis(cPad::AX_UD1, -1.0f);
  if (k == GLUT_KEY_DOWN) gKeyPad->setAxis(cPad::AX_UD1, +1.0f);
}

void SpecialUp(int k, int x, int y)
{
  if (k == GLUT_KEY_LEFT) gKeyPad->setAxis(cPad::AX_LR1, 0.0f);
  if (k == GLUT_KEY_RIGHT) gKeyPad->setAxis(cPad::AX_LR1, 0.0f);
  if (k == GLUT_KEY_UP) gKeyPad->setAxis(cPad::AX_UD1, 0.0f);
  if (k == GLUT_KEY_DOWN) gKeyPad->setAxis(cPad::AX_UD1, 0.0f);
}

//
// GOD begins below!
//

#ifdef _WIN32

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  int    argc = 2;
  char  *argv[] = { "Nothing", "Nothing", "Nothing" };

#else

int main(int argc, char **argv) {

#endif

  width = 640;
  height = 480;

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);  
  glutInitWindowSize(width, height);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("LinWarrior 3D - v. 0.61 public beta");

  glutDisplayFunc(*Draw);
  glutReshapeFunc(*Reshape);
  glutIdleFunc(*Idle);
  
  glutMouseFunc(*MouseFunc);
  glutPassiveMotionFunc(*Mouse);
  glutMotionFunc(*Mouse);
  
  glutKeyboardFunc(*KeyDn);
  glutKeyboardUpFunc(*KeyUp);
  glutSpecialFunc(*SpecialDn);
  glutSpecialUpFunc(*SpecialUp);

  glutSetCursor(GLUT_CURSOR_NONE);
  glutSetKeyRepeat(GLUT_KEY_REPEAT_OFF);

  Init();
  glutMainLoop();

  return 0;
}

