#if !defined(MODELS_INC)
#define MODELS_INC

class cConstruct;
  class cCylinder;
  class cQuader;
    class cCriss;
    class cCross;

class cModel;
  class cPlantModel;
  class cConstructModel;
  class cDummyModel;
  class cParticleModel;

#include "cModel.h"

#include "main.h"

#include <list>
#include <string>

using namespace std;


//-----------------------------------


class cConstruct {
  public:
  public:
  enum               { Construct, Cone, ConeStump, Cylinder, Quader, Criss, Cross };
  public:
  char               mParentName[64];
  char               mName[64];
  int                mType;
  bool               mVisible;
  float              mVolume;
  float              mSurface;
  float              mDamage;
  cSphere            mSphere;
//  cPlane             mPlanes[MAX_PLANES];

  cVector            mPos;
  cVector            mRot;

  float*             mRotator3fv;
  float*             mTranslator3fv;

  cMatrix            mMatrix;
  cMatrix            mInvMatrix;

  bool               mTransformed;
  cConstruct        *mParent;
  public:
                     cConstruct();
  virtual           ~cConstruct() {};
  virtual void       Calculate();
  void               ToWorld();
  virtual void       Draw();

  // Checks for line collision.
  // The line a, b will be transformed using mInvMatrix before checking
  // against the local planes.
  // If there is a collision on the line a, b then true is returned.
  // point will contain the collision point.
  // normal will contain the collision normal of the collision point.
  // dist will contain the distance of the collision point to the point a.
  bool collideLine(float* a, float* b, float* point, float* normal, float* dist);
};


class cCylinder : public cConstruct {
  public:
  float              mRadius;
  float              mHeight;
  public:
  void               Calculate();
  void               Draw();
};


class cQuader : public cConstruct {
  public:
  float              mWidth;
  float              mHeight;
  float              mLength;
  public:
  void               Calculate();
  void               Draw();
};


class cCriss : public cQuader {
  public:
  float              mWidth;
  float              mHeight;
  float              mLength;
  public:
  void               Calculate();
  void               Draw();
};


class cCross : public cQuader {
  public:
  float              mWidth;
  float              mHeight;
  float              mLength;
  public:
  void               Calculate();
  void               Draw();
};


//-----------------------------------


class cConstructModel : public cModel {
  struct model {
    string name;
    list<cConstruct*> constructs;
    bool loadSelf(FILE *f);
  };
  struct context {
    float matrix[16];
    float invmatrix[16];
    float* rotator3fv;
    float* translator3fv;
    model* mod;
    list<float*> rotators;
    list<float*> translators;
    list<float*> matrices;
    list<float*> invmatrices;
  };
  protected:
  list<model*>       mModels;
  protected:
  void*              createContext(char* descriptor);
  void               transformSelf(cEntityModel *entmod);
  void               drawSelf(cEntityModel *entmod);
  cConstruct*        findConstruct(cEntityModel* entmod, char* name, list<float*>::iterator &it);

  public:
  void               setRotator(cEntityModel* entmod, float* rot3fv);
  void               setTranslator(cEntityModel* entmod, float* tra3fv);

  void               setConstructRotator(cEntityModel* entmod, char* name, float* rot3fv);
  void               setConstructTranslator(cEntityModel* entmod, char* name, float* tra3fv);

  float*             getMatrix(cEntityModel* entmod);
  float*             getInvMatrix(cEntityModel* entmod);

  float*             getConstructMatrix(cEntityModel* entmod, char* name);
  float*             getConstructInvMatrix(cEntityModel* entmod, char* name);

//	bool collideLine(cEntityModel *entmod, float* a, float* b, float* point, float* normal, float* dist);
  void collideLine(cEntityModel *entmod, float* a, float* b, cModelCollision *&result);
};


class cPlantModel : public cModel {
  struct context {
    float matrix[16];
    float invmatrix[16];
    float* rotator3fv;
    float* translator3fv;
    list<int> values;
    long displaylist;
  };
  protected:
  void*              createContext(char* descriptor);
  void               transformSelf(cEntityModel *entmod);
  void               drawSelf(cEntityModel *entmod);
  
  public:
  void               setRotator(cEntityModel* entmod, float* rot3fv);
  void               setTranslator(cEntityModel* entmod, float* tra3fv);
};

#define MAX_PARTICLES_PER_SYSTEM 20
class cParticleModel : public cModel {
  struct particle {
    float p[3]; // Position
    float v[3]; // Velocity
    float countdown;
  };
  struct particlesystem { // Context.
    float* sourceMatrixfv;
    float v[3];
    float v_variation[3];
    float countdown;
    list<particle*> particles;
  };
  protected:
  void*              createContext(char* descriptor);
  void               transformSelf(cEntityModel *entmod);
  void               drawSelf(cEntityModel *entmod);

  public:
  void               setSourceMatrix(cEntityModel* entmod, float* matrixfv);
  void               setCountdown(cEntityModel* entmod, float countdown);
  void               spawnParticle(cEntityModel* entmod);
};


class cDummyModel : public cModel {
  struct context {
    float* matrixfv;
  };
  protected:
  void*              createContext(char* descriptor);
  void               transformSelf(cEntityModel *entmod);
  void               drawSelf(cEntityModel *entmod);

  public:
  void               setPlacementMatrix(cEntityModel* entmod, float* matrixfv);
};


#endif
